/*
 * microfsm.c
 *
 * Copyright (C) 2020 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
*/


#include <stdint.h>
#include <microfsm.h>


// Internally-used events
static microfsm_event_t microfsm_reserved_events[] = {
	{(microfsm_signal_t)MICROFSM_SIG_EMPTY},
	{(microfsm_signal_t)MICROFSM_SIG_ENTRY},
	{(microfsm_signal_t)MICROFSM_SIG_EXIT},
};



// API
microfsm_result_t microfsm_init(microfsm_t *me, microfsm_handler_t initial)
{
	microfsm_result_t	r;
	if (me == NULL || initial == NULL)
		return MICROFSM_RC_INVALIDARG;
	me->initial = initial;
	me->state = initial;
	me->userdata = NULL;
	r = (*me->initial)(me, &microfsm_reserved_events[MICROFSM_SIG_ENTRY]);
	// The initial state is allowed to transition on the ENTRY signal
	if (r == MICROFSM_RC_TRANSITION) {
		// signal EXIT on previous state
		(void)(*me->initial)(me, &microfsm_reserved_events[MICROFSM_SIG_EXIT]);
		// signal ENTRY on new state
		(void)(*me->state)(me, &microfsm_reserved_events[MICROFSM_SIG_ENTRY]);
	}
	return MICROFSM_RC_SUCCESS;
}


microfsm_result_t microfsm_dispatch(microfsm_t *me, microfsm_event_t *event)
{
	microfsm_handler_t	s = me->state;
	microfsm_result_t	r = (*s)(me, event);
	if (r == MICROFSM_RC_TRANSITION) {
		// signal EXIT on previous state
		(void)(*s)(me, &microfsm_reserved_events[MICROFSM_SIG_EXIT]);
		// signal ENTRY on new state
		(void)(*me->state)(me, &microfsm_reserved_events[MICROFSM_SIG_ENTRY]);
	}
	return MICROFSM_RC_SUCCESS;
}


microfsm_result_t microfsm_set_userdata(microfsm_t *me, void *userdata)
{
	if (me == NULL)
		return MICROFSM_RC_INVALIDARG;
	me->userdata = userdata;
	return MICROFSM_RC_SUCCESS;
}


