/************************************************************************/
/* pfm.c - Packet Filter Module	Manager source				*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <dlfcn.h>
#include <cutil/slist.h>
#include <helpers.h>
#include <pfm.h>


#define PFM_NAME_SIZE		256


char *pfm_error_strings[] = {
	"Command succeeded",	// 0
	"Module not found",
	"Module version unsupported",
	"Command invalid",
	"Internal module error occurred",
};


// Global structure definition
//pfm_manager_t	pfm_manager;



pfm_manager_t *pfm_open(pfm_init_t *init)
{
	pfm_manager_t	*pfm = NULL;
	pfm = calloc(1, sizeof(pfm_manager_t));
	pfm->provider = init;
	return pfm;
//	(void)memset(&pfm_manager, 0, sizeof(pfm_manager_t));
//	pfm_manager.provider = init;
//	return &pfm_manager;
}



int pfm_set_option(pfm_manager_t *pfm, pfm_option_t option, void *arg)
{
	switch(option) {
		case PFM_OPT_NOOP:
			break;
		case PFM_OPT_SET_LOG_CALLBACK:
			if (arg != NULL)
				pfm->log_callback = arg;
			else
				return EINVAL;
			break;
		default:
			return EINVAL;
	}
	return 0;
}


int pfm_close(pfm_manager_t *pfm)
{
	if (!pfm)
		return EINVAL;
	if (pfm->module_list)
		return 2;
//	free(pfm);
	return 0;
}


pfm_module_t *_pfm_module_find_by_name(pfm_manager_t *manager, char *name)
{
	pfm_module_t    *module = NULL;
	SList		*item = NULL;

	item = manager->module_list;
	while (item) {
		module = slist_get_data(item);
		if (!strcmp(module->info->name, name))
			break;
		item = slist_get_next(item);
	}
	if (!item)
		return NULL;
	return module;
}


int pfm_module_bind_to_port(pfm_manager_t *manager, char *name, uint16_t port)
{
	int		rc;
	pfm_module_t    *module;
	pfm_module_t    *m;
	SList		*item;

	// locate the requested module
	module = _pfm_module_find_by_name(manager, name);
	if (!module)
		return ENOENT;

	// ensure the module is not already bound to the port
	item = manager->port_hash[port];
	if (item) {
		while (item) {
			m = slist_get_data(item);
			if (m == module)
				return EEXIST;
			item = slist_get_next(item);
		}
	}

	// bind the module to the port
	rc = slist_item_add(&manager->port_hash[port], module);

	return rc;
}


int pfm_module_load(pfm_manager_t *manager, char *path)
{
	int		rc;
	int		i;
	pfm_module_t	*module;
//	pfm_t		*(*pfm_module_info)(void);
	pfm_module_info_func	pfm_module_info;

	module = calloc(1, sizeof(pfm_module_t));
	if (!module)
		return errno;
	module->manager = manager;

//	module->dlmodule = dlopen(path, RTLD_NOW|RTLD_DEEPBIND|RTLD_GLOBAL);
	module->dlmodule = dlopen(path, RTLD_NOW|RTLD_GLOBAL);
	if (!module->dlmodule) {
//		printf("DEBUG %s() dlopen() failed\n", __func__);
		return 2;
	}

	pfm_module_info = dlsym(module->dlmodule, "pfm_module_info");
	if (!pfm_module_info) {
//		printf("DEBUG %s() pfm_module_info symbol lookup failed\n", __func__);
		rc = 2;
		goto pfm_module_load_error;
	}
//	rc = (*pfm_module_info)(module->info);
	module->info = dlsym(module->dlmodule, "pfm_info");
	if (!module->info) {
//		printf("DEBUG %s() pfm_info symbol lookup failed\n", __func__);
		rc = 2;
		goto pfm_module_load_error;
	}

	module->pfm_module_open = dlsym(module->dlmodule, "pfm_module_open");
	if (!module->pfm_module_open) {
//		printf("DEBUG %s() pfm_module_open symbol lookup failed\n", __func__);
		rc = 2;
		goto pfm_module_load_error;
	}
	module->pfm_module_close = dlsym(module->dlmodule, "pfm_module_close");
	if (!module->pfm_module_close) {
//		printf("DEBUG %s() pfm_module_close symbol lookup failed\n", __func__);
		rc = 2;
		goto pfm_module_load_error;
	}

	rc = (*module->pfm_module_open)(module);
	if (rc) {
//		printf("DEBUG %s() pfm_module_open() failed\n", __func__);
		goto pfm_module_load_error;
	}

	rc = slist_item_add(&manager->module_list, module);
	if (rc) {
		goto pfm_module_load_error;
	}

	i = 0;
	while (module->info->port_binds[i] != 0) {
		rc = pfm_module_bind_to_port(manager,
			module->info->name,
			module->info->port_binds[i]);
		i++;
	}

//	log_entry("INFO PFM loaded name='%s' version=%d\n",
//		module->info->name,
//		module->info->version);

	return 0;

pfm_module_load_error:
	(void)dlclose(module->dlmodule);
	return rc;
}


int pfm_module_unload(pfm_manager_t *manager, char *name)
{
	int		i;
	int		rc;
	pfm_module_t    *module;

	// locate the requested module
	module = _pfm_module_find_by_name(manager, name);
	if (!module)
		return ENOENT;

	// remove the module from any port binding(s)
	for (i=0 ; i<PFM_HASH_SIZE ; i++) {
		rc = slist_item_remove(&manager->port_hash[i], module);
	}

	// free module resources
	rc = (*module->pfm_module_close)(module);
	if (rc)
		return rc;

	rc = dlclose(module->dlmodule);
	rc = slist_item_remove(&manager->module_list, module);
	free(module);

	return rc;
}


SList *pfm_module_list(pfm_manager_t *pfm)
{
	return pfm->module_list;
}


int pfm_process_packet(pfm_manager_t *manager,
			io_context_t *ctx,
			pnetbuffer_t *in,
			pnetbuffer_t *out)
{
	int		rc;
	SList		*item;
	pfm_module_t	*module;
	pfm_vectors_t	*vectors;

	rc = VOTE_PASS;
	item = manager->port_hash[ctx->dest_port];
	while (item) {
		module = slist_get_data(item);
		if (module->info->version == PFM_VERSION_1) {
			vectors = (pfm_vectors_t*)module->info->vectors;
		} else {
			rc = VOTE_DENY;
			break;
		}
		rc = (*vectors->process_data)(module->info, ctx, in, out);
		if (rc != VOTE_PASS) {
			break;
		}
		item = slist_get_next(item);
	}

	return rc;
}


int pfm_pnetbuffer_add_rc_string(pnetbuffer_t *out, uint16_t err)
{
	uint8_t		rc;
	rc = pnetbuffer_add_uint16(out, err);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_strings(out,
		&pfm_error_strings[err],
		1,
		PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	return rc;
}


int pfm_process_command(pfm_manager_t *manager,
			pnetbuffer_t *in,
			pnetbuffer_t *out)
{
	int		rc;
	uint8_t		len;
	uint8_t		count;
	char		name[PFM_NAME_SIZE];
	pfm_module_t	*module;
	pfm_vectors_t	*vectors;

	// string count
	rc = pnetbuffer_read_uint8(in, &count);
	if (rc != 0) return 2;

	// read module name string
	rc = pnetbuffer_read_uint8(in, &len);
	if (rc != 0) return 2;
	rc = pnetbuffer_read_data(in, name, len);
	if (rc != 0) return 2;
	name[len] = 0;

	// locate module by name
//	printf("searching for module '%s'\n", name);
	module = _pfm_module_find_by_name(manager, name);
	if (!module) {
		rc = pnetbuffer_add_string(out,
			"PFM module not found");
		return rc;
	}
//	printf("located for module '%s'\n", name);

	// pass command to module
	if (module->info->version == PFM_VERSION_1) {
		vectors = (pfm_vectors_t*)module->info->vectors;
	} else {
//		rc = pfm_pnetbuffer_add_rc_string(out, PFM_ERR_MOD_VERSION);
		rc = pnetbuffer_add_string(out,
			"Unsupported PFM version");
		return rc;
	}
	rc = (*vectors->process_command)(module->info, in, out);

	// command results
//	rc = pfm_pnetbuffer_add_rc_string(out, PFM_ERR_SUCCESS);
	return rc;
}



// vector functions for modules to use to communicate back to PFM subsystem
int pfm_vector_ioctl(pfm_module_t *module, int option, void *optionarg)
{
	int				rc = 0;
	pfm_manager_t			*manager;
	ioctl_opt_register_fd_t		*o_rfd;
	if (!module)
		return EINVAL;
	manager = (pfm_manager_t *)module->manager;
	if (!manager)
		return EINVAL;
	switch (option) {
		case PFM_IOCTL_REGISTER_BACKEND_FD:
			// register a backend FD with the I/O loop
			o_rfd = (ioctl_opt_register_fd_t *)optionarg;
//			printf("DEBUG %s() o_rfd->callback=%p\n", __func__, o_rfd->callback);
			rc = ioloop_register_fd(manager->provider->ioloop,
						o_rfd->fd,
						CD_TYPE_BACKEND,
						o_rfd->port,
						o_rfd->callback,
						o_rfd->callback_arg);
			break;
		case PFM_IOCTL_UNREGISTER_BACKEND_FD:
			// unregister a backend FD with the I/O loop
			o_rfd = (ioctl_opt_register_fd_t *)optionarg;
//			printf("DEBUG %s() o_rfd->callback=%p\n", __func__, o_rfd->callback);
			rc = ioloop_unregister_fd(manager->provider->ioloop,
						o_rfd->fd);
			break;
		default:
			// invalid option specified
			return EINVAL;
	}
	return rc;
}

int pfm_vector_log(pfm_module_t *module, int log_level, char *format, ...)
{
	int		rc;
	va_list		args;
	pfm_manager_t	*manager;

	manager = module->manager;
	if (!manager->log_callback)
		return 0;

	va_start(args, format);
	rc = (*manager->log_callback)(log_level, format, args);
	va_end(args);

	return rc;
}


