/************************************************************************/
/* filterset.c - Filterobj defines					*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <pcre.h>

#include <filterset.h>


// Function prototypes
int tree_compare_string(const void *key1, const void *key2);
int tree_destroy_string(void *key);
int tree_compare_filter(const void *key1, const void *key2);
int tree_destroy_filter(void *key);


char	*filterset_type_strings[] = {
	"none",
	"substring",
	"regex",
	"tree",
	"cidr4",
};

char	*filterset_action_strings[] = {
	"none",
	"accept",
	"log",
	"reject",
};



filterset_t *filterset_new(char *name,
			filter_type type,
			filter_action action)
{
	int		rc;
	filterset_t	*fs = NULL;
	AVLTree		*avltree;

	fs = calloc(1, sizeof(filterset_t));
	if (!fs) return NULL;
	fs->type = type;
	fs->action = action;
	fs->name = strdup(name);

	switch(fs->type) {
		case F_TYPE_SUBSTRING:
		case F_TYPE_REGEX:
			fs->impl = NULL;
			break;	
		case F_TYPE_AVLTREE:
			rc = avltree_new(&avltree,
					tree_compare_filter,
					tree_destroy_filter);
			if (rc) goto filterset_new_err;
			fs->impl = avltree;
			break;	
		default:
			goto filterset_new_err;
	}
	return fs;

filterset_new_err:
	if (fs) free(fs);
	return NULL;
}


int filterset_destroy(filterset_t *fs)
{
	int		rc;
	void		*data;
	SList		*slist;
	AVLTree		*avltree;

	if (!fs) return EINVAL;
	rc = 0;
	switch(fs->type) {
		case F_TYPE_SUBSTRING:
		case F_TYPE_REGEX:
			slist = fs->impl;
			while (slist) {
				data = slist_get_data(slist);
				// TODO: free internals first
				free(data);
				slist = slist_get_next(slist);
			}
			slist = fs->impl;
			rc = slist_free(&slist);
			fs->impl = NULL;
			break;	
		case F_TYPE_AVLTREE:
			avltree = fs->impl;
			rc = avltree_free(&avltree);
			fs->impl = NULL;
		default:
			break;
	}
	free(fs->name);
	free(fs);
	return rc;
}


filter_t *slist_get_filter_by_name(SList *slist, char *name)
{
	filter_t	*f;
	SList		*item;
	item = slist;
	while (item) {
		f = (filter_t *)slist_get_data(item);
		if (!strcmp(f->match, name))
			return f;
		item = slist_get_next(item);
	}
	return NULL;
}


int filterset_add_entry(filterset_t *fs, char *match)
{
	int		rc;
	int		err_oset;
	char		*err_str;
	AVLTree		*avltree;
	filter_t	*f;

	if (!fs) return EINVAL;
	rc = 0;
	f = calloc(1, sizeof(filter_t));
	if (!f) return ENOMEM;
	f->type = fs->type;
	f->action = fs->action;
	metric_init(&f->hits, "hits", METRIC_TYPE_UINT32);
	f->match = strdup(match);
	if (!f->match) {
		free(f);
		return ENOMEM;
	}
	switch(fs->type) {
		case F_TYPE_SUBSTRING:
			rc = slist_item_add((SList **)&fs->impl, f);
			if (rc) goto filterset_add_entry_err;
			break;
		case F_TYPE_REGEX:
			f->arg = pcre_compile(f->match,
					0,
					(const char **)&err_str,
					&err_oset,
					NULL);
			if (!f->arg) {
				rc = EINVAL;
				// TODO: return the error str/oset
				goto filterset_add_entry_err;
			}
			rc = slist_item_add((SList **)&fs->impl, f);
			if (rc) goto filterset_add_entry_err;
			break;
		case F_TYPE_AVLTREE:
			avltree = fs->impl;
			rc = avltree_insert(avltree, f);
			if (rc) goto filterset_add_entry_err;
			break;
		default:
			break;
	}
	return 0;

filterset_add_entry_err:
	if (f && f->match) free(f->match);
	if (f) free(f);
	return rc;
}


int filterset_remove_entry(filterset_t *fs, char *match)
{
	int		rc;
	AVLTree		*avltree;
	filter_t	*f = NULL;

	if (!fs) return EINVAL;
	rc = 0;
	switch(fs->type) {
		case F_TYPE_SUBSTRING:
			f = slist_get_filter_by_name(fs->impl, match);
			if (!f) return ENOENT;
			rc = slist_item_remove((SList **)&fs->impl, f);
			if (rc) return rc;
			break;
		case F_TYPE_REGEX:
			f = slist_get_filter_by_name(fs->impl, match);
			if (!f) return ENOENT;
			rc = slist_item_remove((SList **)&fs->impl, f);
			if (rc) return rc;
			pcre_free(f->arg);
			break;
		case F_TYPE_AVLTREE:
			avltree = fs->impl;
			f = avltree_find(avltree, match);
			if (!f) return ENOENT;
			rc = avltree_remove(avltree, match);
			return rc; // the filter was already freed
			break;
		default:
			break;
	}
	if (f && f->match) {
		free(f->match);
		free(f);
	}
	return rc;
}


filter_action filterset_test(filterset_t *fs, char *match)
{
	int		rc;
	int		ovector[32];
	SList		*item;
	AVLTree		*avltree;
	filter_t	*f = NULL;

	if (!fs) return F_ACTION_NONE;
	switch(fs->type) {
		case F_TYPE_SUBSTRING:
			item = fs->impl;
			while (item) {
				f = (filter_t *)slist_get_data(item);
				if (strstr(match, f->match)) {
					metric_incr_u32(&f->hits);
					return f->action;
				}
				item = slist_get_next(item);
			}
			break;
		case F_TYPE_REGEX:
			item = fs->impl;
			while (item) {
				f = (filter_t *)slist_get_data(item);
				rc = pcre_exec(f->arg,	// pcre
					NULL,		// extra
					match,		// subject str
					strlen(match),	// length
					0,		// start oset
					0,		// options
					ovector,	// output vector
					32);		// output vector size
				if (rc >= 1) {
					metric_incr_u32(&f->hits);
					return fs->action;
				}
				item = slist_get_next(item);
			}
			break;
		case F_TYPE_AVLTREE:
			avltree = fs->impl;
			f = avltree_find(avltree, match);
			if (f) return f->action;
			break;
		default:
			break;
	}
	return F_ACTION_ACCEPT;
}


int filterset_enumerate_entries(filterset_t *fs,
                filterset_entry_enum_callback cb,
		void *cb_arg)
{
	int		rc;
	SList		*item;
	AVLTree		*avltree;
	filter_t	*f = NULL;

	if (!fs) return EINVAL;
	if (!cb) return EINVAL;
	rc = 0;
	switch(fs->type) {
		case F_TYPE_SUBSTRING:
		case F_TYPE_REGEX:
			item = fs->impl;
			while (item) {
				f = (filter_t *)slist_get_data(item);
				rc = (*cb)(f, cb_arg);
				if (rc) return rc;
				item = slist_get_next(item);
			}
			break;
		case F_TYPE_AVLTREE:
			avltree = fs->impl;
			break;
		default:
			break;
	}
	return rc;
}




int tree_compare_string(const void *key1, const void *key2)
{
	return strcmp(key1, key2);
}


int tree_destroy_string(void *key)
{
	if (key)
		free(key);
	return 0;
}


int tree_compare_filter(const void *key1, const void *key2)
{
	return strcmp(((filter_t *)key1)->match, ((char *)key2));
}


int tree_destroy_filter(void *key)
{
	filter_t *	f;
	if (key) {
		f = key;
		if (f->match) free(f->match);
		free(f);
	}
	return 0;
}


filter_type filterset_string_to_type(char *str)
{
	int		i;
	for (i=0 ; i<_F_TYPE_COUNT ; i++) {
		if (!strcmp(filterset_type_strings[i], str))
			return i;
	}
	return F_TYPE_NONE;
}


filter_action filterset_string_to_action(char *str)
{
	int		i;
	for (i=0 ; i<_F_ACTION_COUNT ; i++) {
		if (!strcmp(filterset_action_strings[i], str))
			return i;
	}
	return F_TYPE_NONE;
}


char *filterset_type_to_string(filter_type type)
{
	if (type > F_TYPE_NONE && type < _F_TYPE_COUNT)
		return filterset_type_strings[type];
	return NULL;
}


char *filterset_action_to_string(filter_action action)
{
	if (action > F_ACTION_NONE && action < _F_ACTION_COUNT)
		return filterset_action_strings[action];
	return NULL;
}



